@@ -1,6 +1,6 @@
 /*** date-core.c -- our universe of dates
  *
- * Copyright (C) 2011-2016 Sebastian Freundt
+ * Copyright (C) 2011-2024 Sebastian Freundt
  *
  * Author:  Sebastian Freundt <freundt@ga-group.nl>
  *
@@ -105,7 +105,7 @@ struct __md_s {
 /* bizda definitions, reference dates */
 static __attribute__((unused)) const char *bizda_ult[] = {"ultimo", "ult"};
 
-/* arithmetics helpers */
+/* arithmetic helpers */
 static inline unsigned int
 __uimod(signed int x, signed int m)
 {
@@ -121,6 +121,12 @@ __uidiv(signed int x, signed int m)
 	return x >= 0 ? res : x % m ? res - 1 : res;
 }
 
+static inline __attribute__((const, pure)) int
+__sgn(signed int x)
+{
+	return (x > 0) - (x < 0);
+}
+
 
 /* helpers from the calendar files, don't define any aspect, so only
  * the helpers should get included */
@@ -251,8 +257,8 @@ dt_get_md(struct dt_d_s that)
 	}
 }
 
-/* too exotic to be public */
-static int
+/* too exotic to be public, nope bug #81 needs it */
+int
 dt_get_wcnt_mon(struct dt_d_s that)
 {
 	if (LIKELY(that.typ == DT_YMCW)) {
@@ -456,6 +462,8 @@ dt_conv_to_daisy(struct dt_d_s that)
 		return __ldn_to_daisy(that.ldn);
 	case DT_JDN:
 		return __jdn_to_daisy(that.jdn);
+	case DT_MDN:
+		return __mdn_to_daisy(that.mdn);
 	case DT_YD:
 		return __yd_to_daisy(that.yd);
 	case DT_DUNK:
@@ -478,6 +486,10 @@ dt_conv_to_ymd(struct dt_d_s that)
 		goto daisy;
 	case DT_LDN:
 		that.daisy = __ldn_to_daisy(that.ldn);
+		goto daisy;
+	case DT_MDN:
+		that.daisy = __mdn_to_daisy(that.mdn);
+		goto daisy;
 	case DT_DAISY:
 	daisy:
 		return __daisy_to_ymd(that.daisy);
@@ -507,6 +519,10 @@ dt_conv_to_ymcw(struct dt_d_s that)
 		goto daisy;
 	case DT_LDN:
 		that.daisy = __ldn_to_daisy(that.ldn);
+		goto daisy;
+	case DT_MDN:
+		that.daisy = __mdn_to_daisy(that.mdn);
+		goto daisy;
 	case DT_DAISY:
 	daisy:
 		return __daisy_to_ymcw(that.daisy);
@@ -563,6 +579,10 @@ dt_conv_to_ywd(struct dt_d_s this)
 		goto daisy;
 	case DT_LDN:
 		this.daisy = __ldn_to_daisy(this.ldn);
+		goto daisy;
+	case DT_MDN:
+		this.daisy = __mdn_to_daisy(this.mdn);
+		goto daisy;
 	case DT_DAISY:
 	daisy:
 		return __daisy_to_ywd(this.daisy);
@@ -591,6 +611,10 @@ dt_conv_to_yd(struct dt_d_s this)
 		goto daisy;
 	case DT_LDN:
 		this.daisy = __ldn_to_daisy(this.ldn);
+		goto daisy;
+	case DT_MDN:
+		this.daisy = __mdn_to_daisy(this.mdn);
+		goto daisy;
 	case DT_DAISY:
 	daisy:
 		return __daisy_to_yd(this.daisy);
@@ -645,7 +669,7 @@ dt_conv_to_yd(struct dt_d_s this)
 
 DEFVAR const char ymd_dflt[] = "%F";
 DEFVAR const char ymcw_dflt[] = "%Y-%m-%c-%w";
-DEFVAR const char ywd_dflt[] = "%rY-W%V-%u";
+DEFVAR const char ywd_dflt[] = "%G-W%V-%u";
 DEFVAR const char yd_dflt[] = "%Y-%D";
 DEFVAR const char daisy_dflt[] = "%d";
 DEFVAR const char bizsi_dflt[] = "%db";
@@ -653,7 +677,7 @@ DEFVAR const char bizda_dflt[] = "%Y-%m-%db";
 
 DEFVAR const char ymddur_dflt[] = "%Y-%0m-%0d";
 DEFVAR const char ymcwdur_dflt[] = "%Y-%0m-%0w-%0d";
-DEFVAR const char ywddur_dflt[] = "%rY-W%0w-%0d";
+DEFVAR const char ywddur_dflt[] = "%G-W%0w-%d";
 DEFVAR const char yddur_dflt[] = "%Y-%0d";
 DEFVAR const char daisydur_dflt[] = "%d";
 DEFVAR const char bizsidur_dflt[] = "%db";
@@ -772,7 +796,7 @@ __trans_ddurfmt(const char **fmt)
 DEFUN struct dt_d_s
 __guess_dtyp(struct strpd_s d)
 {
-	struct dt_d_s res = dt_d_initialiser();
+	struct dt_d_s res = {DT_DUNK};
 
 	if (LIKELY(d.y > 0 && d.c <= 0 && !d.flags.c_wcnt_p && !d.flags.bizda)) {
 		/* nearly all goes to ymd */
@@ -857,8 +881,8 @@ __guess_dtyp(struct strpd_s d)
 DEFUN struct dt_d_s
 dt_strpd(const char *str, const char *fmt, char **ep)
 {
-	struct dt_d_s res = dt_d_initialiser();
-	struct strpd_s d = strpd_initialiser();
+	struct dt_d_s res = {DT_DUNK};
+	struct strpd_s d = {0};
 	const char *sp = str;
 	const char *fp;
 
@@ -931,7 +955,7 @@ out:
 DEFUN size_t
 dt_strfd(char *restrict buf, size_t bsz, const char *fmt, struct dt_d_s that)
 {
-	struct strpd_s d = strpd_initialiser();
+	struct strpd_s d = {0};
 	const char *fp;
 	char *bp;
 	dt_dtyp_t tgttyp;
@@ -1004,7 +1028,11 @@ dt_strfd(char *restrict buf, size_t bsz, const char *fmt, struct dt_d_s that)
 	case DT_LDN:
 		that.typ = DT_DAISY;
 		that.daisy = __ldn_to_daisy(that.ldn);
-		/* FALLTHROUGH */
+		goto daisy_prep;
+	case DT_MDN:
+		that.typ = DT_DAISY;
+		that.daisy = __mdn_to_daisy(that.mdn);
+		goto daisy_prep;
 	case DT_DAISY:
 	daisy_prep:
 		__prep_strfd_daisy(&d, that.daisy);
@@ -1126,7 +1154,7 @@ out:
 DEFUN size_t
 dt_strfddur(char *restrict buf, size_t bsz, const char *fmt, struct dt_ddur_s that)
 {
-	struct strpd_s d = strpd_initialiser();
+	struct strpd_s d = {0};
 	const char *fp;
 	char *bp;
 
@@ -1277,6 +1305,7 @@ dt_date(dt_dtyp_t outtyp)
 	switch ((res.typ = outtyp)) {
 	case DT_YMD:
 	case DT_YMCW:
+	case DT_YWD:
 	case DT_YD: {
 		struct tm tm;
 		ffff_gmtime(&tm, t);
@@ -1309,11 +1338,21 @@ dt_date(dt_dtyp_t outtyp)
 			res.yd.y = tm.tm_year;
 			res.yd.d = tm.tm_yday;
 			break;
+		case DT_YWD:
+			/* use ordinary conversion to ywd */
+			res.typ = DT_YMD;
+			res.ymd.y = tm.tm_year;
+			res.ymd.m = tm.tm_mon;
+			res.ymd.d = tm.tm_mday;
+			res = dt_dconv(DT_YWD, res);
+			break;
+
 		default:
 			break;
 		}
 		break;
 	}
+
 	case DT_DAISY:
 		/* time_t's base is 1970-01-01, which is daisy 19359 */
 		res.daisy = t / 86400 + 19359;
@@ -1330,7 +1369,10 @@ dt_date(dt_dtyp_t outtyp)
 DEFUN struct dt_d_s
 dt_dconv(dt_dtyp_t tgttyp, struct dt_d_s d)
 {
-	struct dt_d_s res = dt_d_initialiser();
+	struct dt_d_s res = {DT_DUNK};
+
+	/* fix up before conversion */
+	d = dt_dfixup(d);
 
 	switch ((res.typ = tgttyp)) {
 	case DT_YMD:
@@ -1341,7 +1383,8 @@ dt_dconv(dt_dtyp_t tgttyp, struct dt_d_s d)
 		break;
 	case DT_DAISY:
 	case DT_JDN:
-	case DT_LDN: {
+	case DT_LDN:
+	case DT_MDN: {
 		dt_daisy_t tmp = dt_conv_to_daisy(d);
 		switch (tgttyp) {
 		case DT_DAISY:
@@ -1353,6 +1396,9 @@ dt_dconv(dt_dtyp_t tgttyp, struct dt_d_s d)
 		case DT_JDN:
 			res.jdn = __daisy_to_jdn(tmp);
 			break;
+		case DT_MDN:
+			res.mdn = __daisy_to_mdn(tmp);
+			break;
 		default:
 			/* nice one gcc */
 			;
@@ -1381,6 +1427,10 @@ DEFUN struct dt_d_s
 dt_dadd_d(struct dt_d_s d, int n)
 {
 /* add N (gregorian) days to D */
+	if (UNLIKELY(!n)) {
+		/* can't use short-cut return here, it'd upset the IPO/LTO */
+		goto out;
+	}
 	switch (d.typ) {
 	case DT_JDN:
 		d.daisy = __jdn_to_daisy(d.jdn);
@@ -1390,6 +1440,10 @@ dt_dadd_d(struct dt_d_s d, int n)
 		d.daisy = __ldn_to_daisy(d.ldn);
 		goto daisy_add_d;
 
+	case DT_MDN:
+		d.daisy = __mdn_to_daisy(d.mdn);
+		goto daisy_add_d;
+
 	case DT_DAISY:
 	daisy_add_d:
 		d.daisy = __daisy_add_d(d.daisy, n);
@@ -1405,6 +1459,9 @@ dt_dadd_d(struct dt_d_s d, int n)
 		case DT_JDN:
 			d.jdn = __daisy_to_jdn(d.daisy);
 			break;
+		case DT_MDN:
+			d.mdn = __daisy_to_mdn(d.daisy);
+			break;
 		}
 		break;
 
@@ -1434,6 +1491,7 @@ dt_dadd_d(struct dt_d_s d, int n)
 		d.u = 0;
 		break;
 	}
+out:
 	return d;
 }
 
@@ -1441,6 +1499,10 @@ DEFUN struct dt_d_s
 dt_dadd_b(struct dt_d_s d, int n)
 {
 /* add N business days to D */
+	if (UNLIKELY(!n)) {
+		/* cacn't use short-cut return here, it'd upset the IPO/LTO */
+		goto out;
+	}
 	switch (d.typ) {
 	case DT_JDN:
 		d.daisy = __jdn_to_daisy(d.jdn);
@@ -1450,6 +1512,10 @@ dt_dadd_b(struct dt_d_s d, int n)
 		d.daisy = __ldn_to_daisy(d.ldn);
 		goto daisy_add_b;
 
+	case DT_MDN:
+		d.daisy = __mdn_to_daisy(d.mdn);
+		goto daisy_add_b;
+
 	case DT_DAISY:
 	daisy_add_b:
 		d.daisy = __daisy_add_b(d.daisy, n);
@@ -1465,6 +1531,9 @@ dt_dadd_b(struct dt_d_s d, int n)
 		case DT_JDN:
 			d.jdn = __daisy_to_jdn(d.daisy);
 			break;
+		case DT_MDN:
+			d.mdn = __daisy_to_mdn(d.daisy);
+			break;
 		}
 		break;
 
@@ -1494,6 +1563,7 @@ dt_dadd_b(struct dt_d_s d, int n)
 		d.u = 0;
 		break;
 	}
+out:
 	return d;
 }
 
@@ -1501,6 +1571,10 @@ DEFUN struct dt_d_s
 dt_dadd_w(struct dt_d_s d, int n)
 {
 /* add N weeks to D */
+	if (UNLIKELY(!n)) {
+		/* cacn't use short-cut return here, it'd upset the IPO/LTO */
+		goto out;
+	}
 	switch (d.typ) {
 	case DT_JDN:
 		d.daisy = __jdn_to_daisy(d.jdn);
@@ -1510,6 +1584,10 @@ dt_dadd_w(struct dt_d_s d, int n)
 		d.daisy = __ldn_to_daisy(d.ldn);
 		goto daisy_add_w;
 
+	case DT_MDN:
+		d.daisy = __mdn_to_daisy(d.mdn);
+		goto daisy_add_w;
+
 	case DT_DAISY:
 	daisy_add_w:
 		d.daisy = __daisy_add_w(d.daisy, n);
@@ -1525,6 +1603,9 @@ dt_dadd_w(struct dt_d_s d, int n)
 		case DT_JDN:
 			d.jdn = __daisy_to_jdn(d.daisy);
 			break;
+		case DT_MDN:
+			d.mdn = __daisy_to_mdn(d.daisy);
+			break;
 		}
 		break;
 
@@ -1554,6 +1635,7 @@ dt_dadd_w(struct dt_d_s d, int n)
 		d.u = 0;
 		break;
 	}
+out:
 	return d;
 }
 
@@ -1561,9 +1643,13 @@ DEFUN struct dt_d_s
 dt_dadd_m(struct dt_d_s d, int n)
 {
 /* add N months to D */
+	if (UNLIKELY(!n)) {
+		goto out;
+	}
 	switch (d.typ) {
 	case DT_LDN:
 	case DT_JDN:
+	case DT_MDN:
 	case DT_DAISY:
 		/* daisy objects have no notion of months */
 		break;
@@ -1594,6 +1680,7 @@ dt_dadd_m(struct dt_d_s d, int n)
 		d.u = 0;
 		break;
 	}
+out:
 	return d;
 }
 
@@ -1601,9 +1688,14 @@ DEFUN struct dt_d_s
 dt_dadd_y(struct dt_d_s d, int n)
 {
 /* add N years to D */
+	if (UNLIKELY(!n)) {
+		/* cacn't use short-cut return here, it'd upset the IPO/LTO */
+		goto out;
+	}
 	switch (d.typ) {
 	case DT_LDN:
 	case DT_JDN:
+	case DT_MDN:
 	case DT_DAISY:
 		/* daisy objects have no notion of years */
 		break;
@@ -1634,12 +1726,17 @@ dt_dadd_y(struct dt_d_s d, int n)
 		d.u = 0;
 		break;
 	}
+out:
 	return d;
 }
 
 DEFUN struct dt_d_s
 dt_dadd(struct dt_d_s d, struct dt_ddur_s dur)
 {
+	if (UNLIKELY(!dur.dv)) {
+		/* can't use short-cut return here, it'd upset the IPO/LTO */
+		goto out;
+	}
 	switch (dur.durtyp) {
 	case DT_DURD:
 		d = dt_dadd_d(d, dur.dv);
@@ -1650,6 +1747,9 @@ dt_dadd(struct dt_d_s d, struct dt_ddur_s dur)
 	case DT_DURWK:
 		d = dt_dadd_w(d, dur.dv);
 		break;
+	case DT_DURQU:
+		/* just use the month adder */
+		dur.dv *= 3U;
 	case DT_DURMO:
 		d = dt_dadd_m(d, dur.dv);
 		break;
@@ -1661,6 +1761,7 @@ dt_dadd(struct dt_d_s d, struct dt_ddur_s dur)
 		/* huh? */
 		break;
 	}
+out:
 	return d;
 }
 
@@ -1702,8 +1803,11 @@ dt_dur_neg_p(struct dt_ddur_s dur)
 }
 
 DEFUN struct dt_ddur_s
-dt_ddiff(dt_durtyp_t tgttyp, struct dt_d_s d1, struct dt_d_s d2)
+dt_ddiff(dt_durtyp_t tgttyp, struct dt_d_s d1, struct dt_d_s d2, int carry)
 {
+/* carry of 0 doesn't change anything
+ * carry <0 adds -1 to d2 iff d1 <= d2
+ * carry >0 adds 1 to d2 iff d1 >= d2 */
 	struct dt_ddur_s res = dt_make_ddur(DT_DURUNK, 0);
 
 	switch (tgttyp) {
@@ -1711,6 +1815,7 @@ dt_ddiff(dt_durtyp_t tgttyp, struct dt_d_s d1, struct dt_d_s d2)
 	case DT_DURBD: {
 		dt_daisy_t tmp1 = dt_conv_to_daisy(d1);
 		dt_daisy_t tmp2 = dt_conv_to_daisy(d2);
+
 		res = __daisy_diff(tmp1, tmp2);
 
 		/* fix up result in case it's bizsi, i.e. kick weekends */
@@ -1723,25 +1828,61 @@ dt_ddiff(dt_durtyp_t tgttyp, struct dt_d_s d1, struct dt_d_s d2)
 	case DT_DURYMD: {
 		dt_ymd_t tmp1 = dt_conv_to_ymd(d1);
 		dt_ymd_t tmp2 = dt_conv_to_ymd(d2);
+		int fix = __sgn(carry) == -__sgn(tmp2.u - tmp1.u);
+
+		if (UNLIKELY(fix)) {
+			/* add -1 iff tmp2 > tmp1 && carry < 0
+			 * add 1 iff tmp2 < tmp1 && carry > 0, i.e.
+			 * add sgn(carry) iff sgn(tmp2-tmp1) = -sgn(carry) */
+			tmp2 = __ymd_add_d(tmp2, __sgn(carry));
+		}
 		res = __ymd_diff(tmp1, tmp2);
+		res.fix = fix;
 		break;
 	}
 	case DT_DURYMCW: {
 		dt_ymcw_t tmp1 = dt_conv_to_ymcw(d1);
 		dt_ymcw_t tmp2 = dt_conv_to_ymcw(d2);
+		int fix = carry && __sgn(carry) == -__ymcw_cmp(tmp1, tmp2);
+
+		if (UNLIKELY(fix)) {
+			/* add -1 iff tmp2 > tmp1 && carry < 0
+			 * add 1 iff tmp2 < tmp1 && carry > 0, i.e.
+			 * add sgn(carry) iff sgn(tmp2-tmp1) = -sgn(carry) */
+			tmp2 = __ymcw_add_d(tmp2, __sgn(carry));
+		}
 		res = __ymcw_diff(tmp1, tmp2);
+		res.fix = fix;
 		break;
 	}
 	case DT_DURYD: {
 		dt_yd_t tmp1 = dt_conv_to_yd(d1);
 		dt_yd_t tmp2 = dt_conv_to_yd(d2);
+		int fix = __sgn(carry) == -__sgn(tmp2.u - tmp1.u);
+
+		if (UNLIKELY(fix)) {
+			/* add -1 iff tmp2 > tmp1 && carry < 0
+			 * add 1 iff tmp2 < tmp1 && carry > 0, i.e.
+			 * add sgn(carry) iff sgn(tmp2-tmp1) = -sgn(carry) */
+			tmp2 = __yd_add_d(tmp2, __sgn(carry));
+		}
 		res = __yd_diff(tmp1, tmp2);
+		res.fix = fix;
 		break;
 	}
 	case DT_DURYWD: {
 		dt_ywd_t tmp1 = dt_conv_to_ywd(d1);
 		dt_ywd_t tmp2 = dt_conv_to_ywd(d2);
+		int fix = __sgn(carry) == -__sgn(tmp2.u - tmp1.u);
+
+		if (UNLIKELY(fix)) {
+			/* add -1 iff tmp2 > tmp1 && carry < 0
+			 * add 1 iff tmp2 < tmp1 && carry > 0, i.e.
+			 * add sgn(carry) iff sgn(tmp2-tmp1) = -sgn(carry) */
+			tmp2 = __ywd_add_d(tmp2, __sgn(carry));
+		}
 		res = __ywd_diff(tmp1, tmp2);
+		res.fix = fix;
 		break;
 	}
 	case DT_DURBIZDA:
@@ -1787,7 +1928,50 @@ dt_dcmp(struct dt_d_s d1, struct dt_d_s d2)
 DEFUN int
 dt_d_in_range_p(struct dt_d_s d, struct dt_d_s d1, struct dt_d_s d2)
 {
-	return dt_dcmp(d, d1) >= 0 && dt_dcmp(d, d2) <= 0;
+/* use the following multiplication table
+ *
+ * |d,d2|v |d,d1|>  -2 -1  0  1
+ *      -2          -1 -1 -1 -1
+ *      -1          -1  0  1  1
+ *       0          -1  0  1  1
+ *       1          -1  0  0  0
+ *
+ * encoded in a 32bit uint */
+	static const uint32_t m = 0b10010111100101111010101111111111U;
+	const unsigned int i = (dt_dcmp(d, d1) + 2) & 0b11U;
+	const unsigned int j = (dt_dcmp(d, d2) + 2) & 0b11U;
+	return 2 - ((m >> (i * 8U + j * 2U)) & 0b11U);
+}
+
+DEFUN __attribute__((const)) struct dt_d_s
+dt_dfixup(struct dt_d_s d)
+{
+	switch (d.typ) {
+	case DT_YMD:
+		d.ymd = __ymd_fixup(d.ymd);
+		break;
+	case DT_YMCW:
+		d.ymcw = __ymcw_fixup(d.ymcw);
+		break;
+	case DT_YWD:
+		d.ywd = __ywd_fixup(d.ywd);
+		break;
+	case DT_YD:
+		d.yd = __yd_fixup(d.yd);
+		break;
+	case DT_BIZDA:
+		d.bizda = __bizda_fixup(d.bizda);
+		break;
+
+		/* these can't be buggered */
+	case DT_DAISY:
+	case DT_JDN:
+	case DT_LDN:
+	case DT_MDN:
+	default:
+		break;
+	}
+	return d;
 }
 
 #endif	/* INCLUDED_date_core_c_ */
