@@ -1,6 +1,6 @@
 /*** ymd.c -- guts for ymd dates
  *
- * Copyright (C) 2010-2016 Sebastian Freundt
+ * Copyright (C) 2010-2024 Sebastian Freundt
  *
  * Author:  Sebastian Freundt <freundt@ga-group.nl>
  *
@@ -76,7 +76,7 @@ __get_mdays(unsigned int y, unsigned int m)
 	return res - __md_get_yday(y, m, 0);
 }
 
-static __attribute__((pure)) dt_ymd_t
+DEFUN __attribute__((pure)) dt_ymd_t
 __ymd_fixup(dt_ymd_t d)
 {
 /* given dates like 2012-02-32 this returns 2012-02-29 */
@@ -504,6 +504,28 @@ static dt_daisy_t
 __ymd_to_daisy(dt_ymd_t d)
 {
 	dt_daisy_t res;
+#if 1
+/* cassio neri eaf */
+#define s	(82U)
+#define K	(584693U + 146097U * s)
+#define L	(400U * s)
+	const unsigned int J = d.m <= 2;
+	const unsigned int Y = (uint32_t)(d.y + L) - J;
+	const unsigned int M = J ? d.m + 12U : d.m;
+	const unsigned int D = d.d - 1U;
+	const unsigned int C = Y / 100U;
+
+	/* Rata die.*/
+	const unsigned int y_star = 1461U * Y / 4U - C + C / 4U;
+	const unsigned int m_star = (979U * M - 2919U) / 32U;
+	res = y_star + m_star + D;
+
+	/* shift */
+	res -= K;
+#undef s
+#undef K
+#undef L
+#else
 	unsigned int sy = d.y;
 	unsigned int sm = d.m;
 	unsigned int sd;
@@ -526,6 +548,7 @@ __ymd_to_daisy(dt_ymd_t d)
 
 	res = __jan00_daisy(sy);
 	res += __md_get_yday(sy, sm, sd);
+#endif
 	return res;
 }
 
@@ -552,6 +575,7 @@ __ymd_fixup_d(unsigned int y, signed int m, signed int d)
 {
 	dt_ymd_t res = {0};
 
+	m += !m;
 	if (LIKELY(d >= 1 && d <= 28)) {
 		/* all months in our design range have at least 28 days */
 		;
@@ -589,7 +613,9 @@ static dt_ymd_t
 __ymd_add_d(dt_ymd_t d, int n)
 {
 /* add N days to D */
-	signed int tgtd = d.d + n;
+	signed int tgtd;
+	d = __ymd_fixup(d);
+	tgtd = d.d + n + (n < 0 && !d.d);
 
 	/* fixup the day */
 	return __ymd_fixup_d(d.y, d.m, tgtd);
@@ -599,8 +625,13 @@ static dt_ymd_t
 __ymd_add_b(dt_ymd_t d, int n)
 {
 /* add N business days to D */
-	dt_dow_t wd = __ymd_get_wday(d);
-	int tgtd = d.d + __get_d_equiv(wd, n);
+	dt_dow_t wd;
+	signed int tgtd;
+
+	d = __ymd_fixup(d);
+	d.d ^= n < 0 && !d.d;
+	wd = __ymd_get_wday(d);
+	tgtd = d.d + __get_d_equiv(wd, n);
 
 	/* fixup the day, i.e. 2012-01-34 -> 2012-02-03 */
 	return __ymd_fixup_d(d.y, d.m, tgtd);
@@ -628,8 +659,8 @@ __ymd_add_m(dt_ymd_t d, int n)
 		--d.y;
 	}
 	/* final assignment */
-	d.m = tgtm;
-	return __ymd_fixup(d);
+	d.m = tgtm + (n < 0 && !d.m);
+	return d;
 }
 
 static dt_ymd_t
@@ -637,7 +668,7 @@ __ymd_add_y(dt_ymd_t d, int n)
 {
 /* add N years to D */
 	d.y += n;
-	return __ymd_fixup(d);
+	return d;
 }
 #endif	/* ASPECT_ADD */
 
@@ -672,8 +703,18 @@ __ymd_diff(dt_ymd_t d1, dt_ymd_t d2)
 			d2m = GREG_MONTHS_P_YEAR;
 			d2y--;
 		}
-		tgtd += __get_mdays(d2y, d2m);
 		tgtm--;
+		if (UNLIKELY((tgtd += __get_mdays(d2y, d2m)) < 0)) {
+			/* super special case when d2m is feb and the
+			 * difference is >= 28 or 29,
+			 * unrolled because it cannot happen again */
+			if (--d2m < 1) {
+				d2m = GREG_MONTHS_P_YEAR;
+				d2y--;
+			}
+			tgtm--;
+			tgtd += __get_mdays(d2y, d2m);
+		}
 #if !defined WITH_FAST_ARITH || defined OMIT_FIXUPS
 		/* the non-fast arith has done the fixup already */
 #else  /* WITH_FAST_ARITH && !defined OMIT_FIXUPS */
